!Module for the padding of pressure source and 2D interpolation with the adjoint such that truncating and 2D interpolating
!Model dimension (z,x,y)
!Data dimension (1)
!D(1) = Add_source2D(Model(z,x,y))
!Model(z,x,y) = Add_source2D*(D(1))
module Add_source2D_pressure

  implicit none
  !Module variables
  real, private									:: x, y, z, x0, z0, dx, dz !position of the point and origin of the axes
  integer, private								:: nz,nx, ny, nt !number of time samples
  real, dimension(:,:,:),allocatable,private	:: bulk

  contains
	
	!Initialization routine
	subroutine Add_source2D_pressure_init(nz_in, nx_in, ny_in, nt_in, z_in, x_in, y_in, z0_in, x0_in, dz_in, dx_in)
		
		!Input variables	
		real, intent(in)				:: z_in, x_in, y_in, z0_in, x0_in, dz_in, dx_in
		integer, intent(in)				:: nz_in, nx_in, ny_in, nt_in

		!Assigning the internal variables
		x = x_in
		y = y_in
		z = z_in
		x0 = x0_in
		z0 = z0_in
		dx = dx_in
		dz = dz_in
		nx = nx_in
		ny = ny_in
		nz = nz_in
		nt = nt_in

		return
	end subroutine Add_source2D_pressure_init
	
	subroutine Add_source2D_pressure_init2(lame_in,mu_in,nz_in, nx_in, ny_in)
		real, dimension(:,:,:), pointer			:: lame_in, mu_in
		integer, intent(in)						:: nz_in, nx_in, ny_in
		
		if(.not. allocated(bulk)) allocate(bulk(nz_in,nx_in,ny_in))
		
		bulk = lame_in + 2.0/3.0*mu_in
	
	end subroutine Add_source2D_pressure_init2

  	subroutine Add_source2D_pressurez_call(adj,add,mm,dd)
    	logical,intent(in)   	:: adj,add
   		real,dimension(:)  		:: mm
   		real			  		:: dd
    		call Add_source2D_pressurez_op(adj,add,mm,dd)
  	end subroutine Add_source2D_pressurez_call

	
	!Operator routine
	subroutine Add_source2D_pressurez_op(adj, add, mm, dd)
	
		!Input-Output variables
		logical, intent(in)								:: adj, add
		real, dimension(nz,nx,ny), intent(inout)		:: mm !Padded data (z,x,y)
		real, intent(inout)								:: dd !Data(1)
		!Internal variables
		integer											:: zi, xi
		real											:: zf, xf, wz, wx, scale_bulk
		
		integer::it


		xf = 1. + (x-x0)/dx; xi = xf; wx = xf - xi
		zf = 1. + (z-z0)/dz; zi = zf; wz = zf - zi
		scale_bulk = -(bulk(zi,xi,1)*(1.-wz)*(1.-wx) + &
					 bulk(zi,xi+1,1)*(1.-wz)*(wx)   + &
					 bulk(zi+1,xi,1)*(wz)*(1.-wx)   + &
					 bulk(zi+1,xi+1,1)*(wz)*(wx))*0.5
    	if(adj) then !Adjoint
        		if (.not. add) mm = 0. !Zero-out the model
				mm(zi+1,xi,1)       = mm(zi+1,xi,1)       + dd*(1.-wz)*(1.-wx)*scale_bulk/dz
				mm(zi+1,xi+1,1)     = mm(zi+1,xi+1,1)     + dd*(1.-wz)*(wx)*scale_bulk/dz
				mm(zi+2,xi,1)     = mm(zi+2,xi,1)     	  + dd*(wz)*(1.-wx)*scale_bulk/dz
				mm(zi+2,xi+1,1)   = mm(zi+2,xi+1,1)   	  + dd*(wz)*(wx)*scale_bulk/dz
				
				mm(zi-1,xi,1)       = mm(zi-1,xi,1)       - dd*(1.-wz)*(1.-wx)*scale_bulk/dz
				mm(zi-1,xi+1,1)     = mm(zi-1,xi+1,1)     - dd*(1.-wz)*(wx)*scale_bulk/dz
				mm(zi,xi,1)     = mm(zi,xi,1)     		  - dd*(wz)*(1.-wx)*scale_bulk/dz
				mm(zi,xi+1,1)   = mm(zi,xi+1,1)   		  - dd*(wz)*(wx)*scale_bulk/dz
			

		else	     !Forward
			if(.not. add) dd = 0. !Zero-out the data			
				dd = dd + scale_bulk*((mm(zi+1,xi,1)*(1.-wz)*(1.-wx) 	+ &
					 	        mm(zi+1,xi+1,1)*(1.-wz)*(wx)+ &
						  	    mm(zi+2,xi,1)*(wz)*(1.-wx)  + &
					  	  		mm(zi+2,xi+1,1)*(wz)*(wx) 		+ &
					  	  	 -(mm(zi-1,xi,1)*(1.-wz)*(1.-wx)+ &
					 	        mm(zi-1,xi+1,1)*(1.-wz)*(wx)+ &
						  	    mm(zi,xi,1)*(wz)*(1.-wx)  	+ &
					  	  		mm(zi,xi+1,1)*(wz)*(wx) ))/dz)	
		endif
		
		return
	end subroutine Add_source2D_pressurez_op
	
	
	subroutine Add_source2D_pressurex_call(adj,add,mm,dd)
    	logical,intent(in)   	:: adj,add
   		real,dimension(:)  		:: mm
   		real			  		:: dd
    		call Add_source2D_pressurex_op(adj,add,mm,dd)
  	end subroutine Add_source2D_pressurex_call
	
	
	!Operator routine
	subroutine Add_source2D_pressurex_op(adj, add, mm, dd)
	
		!Input-Output variables
		logical, intent(in)								:: adj, add
		real, dimension(nz,nx,ny), intent(inout)		:: mm !Padded data (z,x,y)
		real, intent(inout)								:: dd !Data(1)
		!Internal variables
		integer											:: zi, xi
		real											:: zf, xf, wz, wx, scale_bulk
		
		integer::it


		xf = 1. + (x-x0)/dx; xi = xf; wx = xf - xi
		zf = 1. + (z-z0)/dz; zi = zf; wz = zf - zi
		scale_bulk = (bulk(zi,xi,1)*(1.-wz)*(1.-wx) + &
					 bulk(zi,xi+1,1)*(1.-wz)*(wx)   + &
					 bulk(zi+1,xi,1)*(wz)*(1.-wx)   + &
					 bulk(zi+1,xi+1,1)*(wz)*(wx))*0.5
    	if(adj) then !Adjoint
        		if (.not. add) mm = 0. !Zero-out the model
				
				mm(zi,xi+1,1)       = mm(zi,xi+1,1)       + dd*(1.-wz)*(1.-wx)*scale_bulk/dx
				mm(zi,xi+2,1)     	= mm(zi,xi+2,1)       + dd*(1.-wz)*(wx)*scale_bulk/dx
				mm(zi+1,xi+1,1)     = mm(zi+1,xi+1,1)     + dd*(wz)*(1.-wx)*scale_bulk/dx
				mm(zi+1,xi+2,1)   	= mm(zi+1,xi+2,1)  	  + dd*(wz)*(wx)*scale_bulk/dx
				
				mm(zi,xi-1,1)       = mm(zi,xi-1,1)       - dd*(1.-wz)*(1.-wx)*scale_bulk/dx
				mm(zi,xi,1)     	= mm(zi,xi,1)     	  - dd*(1.-wz)*(wx)*scale_bulk/dx
				mm(zi+1,xi-1,1)     = mm(zi+1,xi-1,1)     - dd*(wz)*(1.-wx)*scale_bulk/dx
				mm(zi+1,xi+1,1)  	= mm(zi+1,xi,1)   	  - dd*(wz)*(wx)*scale_bulk/dx
			

		else	     !Forward
			if(.not. add) dd = 0. !Zero-out the data			
				dd = dd + scale_bulk*((mm(zi,xi+1,1)*(1.-wz)*(1.-wx) + &
					 	       		   mm(zi,xi+2,1)*(1.-wz)*(wx)	 + &
						  	    	   mm(zi+1,xi+1,1)*(wz)*(1.-wx)  + &
					  	  			   mm(zi+1,xi+2,1)*(wz)*(wx) 	 + &
					  	  	  		 -(mm(zi,xi-1,1)*(1.-wz)*(1.-wx) + &
					 	        	   mm(zi,xi,1)*(1.-wz)*(wx)		 + &
						  	    	   mm(zi+1,xi-1,1)*(wz)*(1.-wx)  + &
					  	  			   mm(zi+1,xi,1)*(wz)*(wx) ))/dx)	
		endif
		
		return
	end subroutine Add_source2D_pressurex_op
	
	
	subroutine Add_source2D_pressure_close()
  		if(allocated(bulk))deallocate(bulk)  		
  		
  	end subroutine  Add_source2D_pressure_close
	

end module Add_source2D_pressure




