!Module for the padding of the source and 3D interpolation with the adjoint such that truncating and 3D interpolating
!Model dimension (z,x,y)
!Data dimension (1)
!D(1) = Add_source3D_dipole(Model(z,x,y))
!Model(z,x,y) = Add_source3D_dipole*(D(1))
module Add_source3D_dipole

  implicit none
  !Module variables
  real, private					:: x, y, z, x0, y0, z0, dx, dy, dz !position of the point and origin of the axes
  integer, private				:: nz,nx, ny, nt !number of time samples

  contains
	
	!Initialization routine
	subroutine Add_source3D_dipole_init(nz_in, nx_in, ny_in, nt_in, z_in, x_in, y_in, z0_in, x0_in, y0_in, dz_in, dx_in, dy_in)
		
		!Input variables	
		real, intent(in)				:: z_in, x_in, y_in, z0_in, x0_in, y0_in, dz_in, dx_in, dy_in
		integer, intent(in)				:: nz_in, nx_in, ny_in, nt_in

		!Assigning the internal variables
		x = x_in
		y = y_in
		z = z_in
		x0 = x0_in
		y0 = y0_in
		z0 = z0_in
		dx = dx_in
		dy = dy_in
		dz = dz_in
		nx = nx_in
		ny = ny_in
		nz = nz_in
		nt = nt_in

		return
	end subroutine Add_source3D_dipole_init

  	subroutine Add_source3D_dipole_call(adj,add,mm,dd)
    	logical,intent(in)   		:: adj,add
   		real,dimension(:)  		:: mm
   		real			  		:: dd
    		call Add_source3D_dipole_op(adj,add,mm,dd)
  	end subroutine Add_source3D_dipole_call

	
	!Operator routine
	subroutine Add_source3D_dipole_op(adj, add, mm, dd)
	
		!Input-Output variables
		logical, intent(in)								:: adj, add
		real, dimension(nz,nx,ny), intent(inout)		:: mm !Padded data (z,x,y)
		real, intent(inout)								:: dd !Data(1)
		!Internal variables
		integer											:: zi, xi, yi
		real											:: zf, xf, yf, wz, wx, wy

		xf = 1. + (x-x0)/dx; xi = xf; wx = xf - xi
		yf = 1. + (y-y0)/dy; yi = yf; wy = yf - yi
		zf = 1. + (z-z0)/dz; zi = zf; wz = zf - zi
    		if(adj) then !Adjoint
        		if (.not. add) mm = 0. !Zero-out the model
				mm(zi,xi,yi)       = mm(zi,xi,yi)       + dd*(1.-wz)*(1.-wx)*(1.-wy)
				mm(zi,xi+1,yi)     = mm(zi,xi+1,yi)     + dd*(1.-wz)*(wx)*(1.-wy)
				mm(zi,xi,yi+1)     = mm(zi,xi,yi+1)     + dd*(1.-wz)*(1.-wx)*(wy)
				mm(zi,xi+1,yi+1)   = mm(zi,xi+1,yi+1)   + dd*(1.-wz)*(wx)*(wy)
				mm(zi+1,xi,yi)     = mm(zi+1,xi,yi)     + dd*(wz)*(1.-wx)*(1.-wy)
				mm(zi+1,xi+1,yi)   = mm(zi+1,xi+1,yi)   + dd*(wz)*(wx)*(1.-wy)
				mm(zi+1,xi,yi+1)   = mm(zi+1,xi,yi+1)   + dd*(wz)*(1.-wx)*(wy)
				mm(zi+1,xi+1,yi+1) = mm(zi+1,xi+1,yi+1) + dd*(wz)*(wx)*(wy)

		else	     !Forward
			if(.not. add) dd = 0. !Zero-out the data
				dd = dd + mm(zi,xi,yi)*(1.-wz)*(1.-wx)*(1.-wy) + &
					 	  mm(zi,xi+1,yi)*(1.-wz)*(wx)*(1.-wy)  + &
						  mm(zi,xi,yi+1)*(1.-wz)*(1.-wx)*(wy)  + &
						  mm(zi,xi+1,yi+1)*(1.-wz)*(wx)*(wy)   + &
						  mm(zi+1,xi,yi)*(wz)*(1.-wx)*(1.-wy)  + &
					  	  mm(zi+1,xi+1,yi)*(wz)*(wx)*(1.-wy)   + &
						  mm(zi+1,xi,yi+1)*(wz)*(1.-wx)*(wy)   + &
						  mm(zi+1,xi+1,yi+1)*(wz)*(wx)*(wy)
		endif
		
		return
	end subroutine Add_source3D_dipole_op
	

end module Add_source3D_dipole




