!Module for computing the C2 operator (see Ali's note) (v**2dt**2)laplacian+2I
!Model dimension (z,x,y) Input wavefield
!Data dimension (z,x,y) Ouput wavefield
!D(z,x,y) = C2(Model(z,x,y))
!Model(z,x,y) = C2*(D(z,x,y))
module C2

  use Laplacian

  implicit none
  !Module variables
  integer, private								:: nz, nx, ny
  !real, dimension(:,:,:), pointer, private		:: vel_module
  real, private									:: dt_module, dt2
  real, dimension(:,:,:),allocatable, private	:: lapla_tmp, scale, tmp

  contains


	!Initialization routine
	subroutine C2_init(nz_in, nx_in, ny_in, vel_in, dt_in, dz_in, dx_in, dy_in, twoD_in)
		
		!Input variables	
		integer, intent(in)						:: nz_in, nx_in, ny_in
		real, dimension(:,:,:), pointer			:: vel_in
		real									:: dt_in, dz_in, dx_in, dy_in
		logical, intent(in)						:: twoD_in
		!Internal variables

		!Assigning the internal variables
		nz = nz_in
		nx = nx_in
		ny = ny_in
		!if (.not. associated(vel_module)) allocate(vel_module(nz,nx,ny))
		!vel_module => vel_in
		dt_module = dt_in
		dt2 = dt_module*dt_module
		if (.not. allocated(lapla_tmp)) allocate(lapla_tmp(nz,nx,ny))
		if (.not. allocated(tmp)) allocate(tmp(nz,nx,ny))
		lapla_tmp = 0.; tmp = 0.
		if (.not. allocated(scale)) allocate(scale(nz,nx,ny))
		scale = vel_in(:,:,:)*vel_in(:,:,:)*dt2

		!Initializing other modules
		call Laplacian_init(nz, nx, ny, dz_in, dx_in, dy_in, twoD_in)

		return
	end subroutine C2_init


  	subroutine C2_call(adj,add,mm,dd)
    	logical,intent(in)   :: adj,add
   		real,dimension(:)  	 :: mm,dd
    		call C2_op(adj,add,mm,dd)
  	end subroutine C2_call

	
	!Operator routine
	subroutine C2_op(adj, add, mm, dd)
	
		!Input-Output variables
		logical, intent(in)					:: adj, add
		real, dimension(nz,nx,ny)			:: mm 
		real, dimension(nz,nx,ny)			:: dd 
		!Internal variables
		integer								:: it

		lapla_tmp = 0.

    	if(adj) then !Adjoint
        	if (.not. add) mm = 0. !Zero-out the model
			!call Laplacian_op(.true.,.false.,lapla_tmp,vel_module(:,:,:)*vel_module(:,:,:)*dt2*(dd(:,:,:)))
			tmp(:,:,:) = scale(:,:,:)*dd(:,:,:)
			call Laplacian_op(.true.,.false.,lapla_tmp,tmp)
			mm(:,:,:) = mm(:,:,:) + lapla_tmp(:,:,:) + 2.*dd(:,:,:)	

		else	     !Forward
			if(.not. add) dd = 0. !Zero-out the data
			call Laplacian_op(.false.,.false.,mm(:,:,:),lapla_tmp)
			!dd(:,:,:) = dd(:,:,:) + vel_module(:,:,:)*vel_module(:,:,:)*dt2*(lapla_tmp(:,:,:)) + 2.*mm(:,:,:)
			dd(:,:,:) = dd(:,:,:) + scale(:,:,:)*(lapla_tmp(:,:,:)) + 2.*mm(:,:,:)
		endif
		
		return
	end subroutine C2_op

	!Subroutine for deallocating variables
	subroutine C2_close()
		if (allocated(lapla_tmp)) deallocate(lapla_tmp)
		if (allocated(tmp)) deallocate(tmp)
		if (allocated(scale)) deallocate(scale)
	end subroutine C2_close


end module C2
