!Module for computing the C2_acoustic operator (Kdt**2)laplacian+2I
!Model dimension (z,x,y) Input wavefield
!Data dimension (z,x,y) Ouput wavefield
!D(z,x,y) = C2_acoustic(Model(z,x,y))
!Model(z,x,y) = C2_acoustic*(D(z,x,y))
module C2_acoustic

  !use Db2
  use DbD10pt
  !use DbD8pt

  implicit none
  !Module variables
  integer, private								:: nz, nx, ny
  real, private									:: dt_module, dt2
  real, dimension(:,:,:),allocatable, private	:: db2_tmp, scale

  contains


	!Initialization routine
	subroutine C2_acoustic_init(nz_in, nx_in, ny_in, k_in, rho_in, dt_in, dz_in, dx_in, dy_in, twoD_in)
		
		!Input variables	
		integer, intent(in)						:: nz_in, nx_in, ny_in
		real, dimension(:,:,:), pointer			:: k_in, rho_in
		real									:: dt_in, dz_in, dx_in, dy_in
		logical, intent(in)						:: twoD_in
		!Internal variables

		!Assigning the internal variables
		nz = nz_in
		nx = nx_in
		ny = ny_in

		dt_module = dt_in
		dt2 = dt_module*dt_module
		if (.not. allocated(db2_tmp)) allocate(db2_tmp(nz,nx,ny))
		db2_tmp = 0.
		if (.not. allocated(scale)) allocate(scale(nz,nx,ny))
		scale = k_in(:,:,:)*dt2

		!Initializing other modules
		!call Db2_init(nz, nx, ny, dz_in, dx_in, dy_in, rho_in, twoD_in)
		call DbD10pt_init(nz, nx, ny, dz_in, dx_in, dy_in, rho_in, twoD_in)
		!call DbD8pt_init(nz, nx, ny, dz_in, dx_in, dy_in, rho_in, twoD_in)

		return
	end subroutine C2_acoustic_init


  	subroutine C2_acoustic_call(adj,add,mm,dd)
    	logical,intent(in)   :: adj,add
   		real,dimension(:)  	 :: mm,dd
    		call C2_acoustic_op(adj,add,mm,dd)
  	end subroutine C2_acoustic_call

	
	!Operator routine
	subroutine C2_acoustic_op(adj, add, mm, dd)
	
		!Input-Output variables
		logical, intent(in)					:: adj, add
		real, dimension(nz,nx,ny)			:: mm 
		real, dimension(nz,nx,ny)			:: dd 
		!Internal variables
		integer								:: it

		db2_tmp = 0.

    		if(adj) then !Adjoint
        		if (.not. add) mm = 0. !Zero-out the model
			!call Db2_op(.true.,.false.,db2_tmp,scale(:,:,:)*(dd(:,:,:)))
			call DbD10pt_op(.true.,.false.,db2_tmp,scale(:,:,:)*(dd(:,:,:)))
			!call DbD8pt_op(.true.,.false.,db2_tmp,scale(:,:,:)*(dd(:,:,:)))
			mm(:,:,:) = mm(:,:,:) + db2_tmp(:,:,:) + 2.*dd(:,:,:)	

		else	     !Forward
			if(.not. add) dd = 0. !Zero-out the data
			!call Db2_op(.false.,.false.,mm(:,:,:),db2_tmp)
			call DbD10pt_op(.false.,.false.,mm(:,:,:),db2_tmp)
			!call DbD8pt_op(.false.,.false.,mm(:,:,:),db2_tmp)
			dd(:,:,:) = dd(:,:,:) + scale(:,:,:)*(db2_tmp(:,:,:)) + 2.*mm(:,:,:)
		endif
		
		return
	end subroutine C2_acoustic_op

	!Subroutine for deallocating variables
	subroutine C2_acoustic_close()
		if (allocated(db2_tmp)) deallocate(db2_tmp)
		if (allocated(scale)) deallocate(scale)
		!call Db2_close()
		call DbD10pt_close()
		!call DbD8pt_close()
	end subroutine C2_acoustic_close


end module C2_acoustic
