!Module for computing the C5 operator (see Ali's note)
!Model dimension (z,x,y) Source wavefield
!Data dimension (z,x,y) Wavefield
!D(z,x,y) = C5(Model(z,x,y))
!Model(z,x,y) = C5*(D(z,x,y))
module C5

  use C2
  use weight
  use G

  implicit none
  !Module variables
  integer, private									:: nz, nx, ny, nt
  real, private										:: dt
  real, dimension(:,:,:), allocatable, private		:: tmp, tmp1, scale

  contains


	!Initialization routine
	subroutine C5_init(nz_in, nx_in, ny_in, vel_in, dt_in, dz_in, dx_in, dy_in, twoD_in)
		
		!Input variables	
		integer, intent(in)						:: nz_in, nx_in, ny_in
		real, dimension(:,:,:), pointer			:: vel_in
		real									:: dt_in, dz_in, dx_in, dy_in
		logical, intent(in)						:: twoD_in

		!Assigning the internal variables
		nz = nz_in
		nx = nx_in
		ny = ny_in
		dt = dt_in

		!if (.not. associated(vel_inter)) allocate(vel_inter(nz,nx,ny))
		!vel_inter => vel_in

		if (.not. allocated(tmp)) allocate(tmp(nz,nx,ny))
		tmp = 0.
		if (.not. allocated(tmp1)) allocate(tmp1(nz,nx,ny))
		tmp1 = 0.
		if (.not. allocated(scale)) allocate(scale(nz,nx,ny))
		scale = vel_in(:,:,:)*dt
		
		!Initializing other modules
		call C2_init(nz, nx, ny, vel_in, dt_in, dz_in, dx_in, dy_in, twoD_in)
		

		return
	end subroutine C5_init


  	subroutine C5_call(adj,add,mm,dd)
    	logical,intent(in)  :: adj,add
   		real,dimension(:)  	:: mm,dd
    		call C5_op(adj,add,mm,dd)
  	end subroutine C5_call

	
	!Operator routine
	subroutine C5_op(adj, add, mm, dd)
	
		!Input-Output variables	
		logical, intent(in)					:: adj, add
		real, dimension(nz,nx,ny)			:: mm 
		real, dimension(nz,nx,ny)			:: dd 
		!Internal variables

    	if(adj) then !Adjoint
        	if (.not. add) mm = 0. !Zero-out the model
			!Operator: (I-G*v*dt)w
			call weight_op(.true.,.false.,tmp,dd(:,:,:))
			!call G_op(.true.,.false.,tmp1,(tmp(:,:,:)*vel_inter(:,:,:)*dt))
			mm(:,:,:) = mm(:,:,:) + tmp(:,:,:)
			tmp(:,:,:)= tmp(:,:,:)*scale(:,:,:)
			call G_op(.true.,.false.,tmp1,tmp(:,:,:))
			mm(:,:,:) = mm(:,:,:) - tmp1(:,:,:)
			!Operator: (lapla*v^2*dt^2+2I)(I-w)
			call weight_op(.true.,.false.,tmp,dd(:,:,:))
			tmp(:,:,:) = dd(:,:,:) - tmp(:,:,:)
			call C2_op(.true.,.false.,tmp1,tmp)
			mm(:,:,:) = mm(:,:,:) + tmp1(:,:,:)

		else	     !Forward
			if(.not. add) dd = 0. !Zero-out the data
			!Operator: (I-w)(v^2*dt^2*lapla+2I)
			call C2_op(.false.,.false.,mm(:,:,:),tmp)
			call weight_op(.false.,.false.,tmp,tmp1)
			dd(:,:,:) = dd(:,:,:) + tmp(:,:,:) - tmp1(:,:,:)
			!!Operator: w(I-v*dt*G)
			call G_op(.false.,.false.,mm(:,:,:),tmp1(:,:,:))
			!call weight_op(.false.,.false.,(mm(:,:,:)-vel_inter(:,:,:)*dt*tmp1(:,:,:)),tmp(:,:,:))
			tmp1(:,:,:)=(mm(:,:,:)-scale(:,:,:)*tmp1(:,:,:))
			call weight_op(.false.,.false.,tmp1(:,:,:),tmp(:,:,:))
			dd(:,:,:) = dd(:,:,:) + tmp(:,:,:)
		endif

		return
	end subroutine C5_op


	!Subroutine for deallocating variables
	subroutine C5_close()
		if (allocated(tmp)) deallocate(tmp)
		if (allocated(tmp1)) deallocate(tmp1)
		if (allocated(scale)) deallocate(scale)
	end subroutine C5_close


end module C5
