!Program for performing derivative test on DbD10pt module

program DbD10pt_deriv_test


	!Used modules 
	use sep
	use DbD10pt
	
	
	implicit none
	real								:: dz, dx, dy, deriv
	integer								:: nz, nx, ny, iz, ix, iy, ik, z, x, y
	logical								:: twoD
	real, dimension(:,:,:), allocatable	:: rho, mm, dd
	
	call sep_init (SOURCE)
	
	!Reading dimension of the model space (and data)
	call from_par("nz",nz)
	call from_par("nx",nx)
	call from_par("ny",ny)
	call from_par("dz",dz)
	call from_par("dx",dx)
	call from_par("dy",dy)
	
	if(ny .eq. 1) then
		twoD = .true.
	else
		twoD = .false.
	endif
	allocate(rho(nz,nx,ny),mm(nz,nx,ny),dd(nz,nx,ny))
	mm = 0.0
	dd = 0.0
	rho = 1.
	!Since rho=1 we have the operator to be Dm*Dp = D2 (second order derivative 10th order)
	!Input different order polynomial functions based on f(z)k = z(i)^k = (i*dz)^k
	!In 2D we have f(z,x)k = (i*dz + j*dx)^k 
	!In 3D we have f(z,x,y)k = (i*dz + j*dx + l*dy)^k
	if(twoD)then
		do ix = 1, nx
			do iz = 1, nz
				mm(iz,ix,1) = (iz - 1)*dz +(ix - 1)*dx
			enddo
		enddo
	else
		do iy = 1, ny
			do ix = 1, nx
				do iz = 1, nz
					mm(iz,ix,iy) = (iz - 1)*dz +(ix - 1)*dx + +(iy - 1)*dy
				enddo
			enddo
		enddo
	endif	
	
	!Calling the subroutines to perform dot product test
	call DbD10pt_init(nz, nx, ny, dz, dx, dy, rho, twoD)
	z = nz/2+1
	x = nx/2+1
	y = ny/2+1
	
	!Testing the different polynomial functions (from 0th-order to 10th-order)
	do ik = 0, 10
		write(0,*)"#########################################"
		write(0,*)"Testing polynomial of order:",ik,"th"
		write(0,*)"Derivative at",z,"nz",x,"nx",y,"ny:"
		call DbD10pt_op(.false., .false., mm**ik, dd)
		write(0,*)"Computed derivative: ",dd(z,x,y)
		if(twoD)then
			deriv = ik*(ik-1)*((nz/2)*dz+(nx/2)*dx)**(ik-2)*2.0
		else
			deriv = ik*(ik-1)*((nz/2)*dz+(nx/2)*dx+(ny/2)*dy)**(ik-2)*3.0
		endif
		write(0,*)"Derivative should be:", deriv
		write(0,*)"Absolute error:", dd(z,x,y)-deriv
		write(0,*)"Relative error:", (dd(z,x,y)-deriv)/deriv*100.,"%"
		write(0,*)"#########################################"
	enddo

	
	
	call DbD10pt_close()
	deallocate(dd,mm,rho)
	call sep_close()
	call exit (0)
	

end program DbD10pt_deriv_test
