!Module for padding the velocity model for the absorbing boundaries

module Pad_vel

  implicit none
				
  contains
	
	subroutine Pad_velocity(vel,vel_pad,nv_pad,oz,oy,ox,nz_out,nx_out,ny_out,dz,dx,dy,twoD)
		!Input-output variables
	  	real, dimension(:,:,:), allocatable, intent(inout)		:: vel   !velocity model (single precision)
	  	real, dimension(:,:,:), pointer,intent(inout)			:: vel_pad !Padded velocity model
  		integer, intent(in)										:: nv_pad !Padding for the velocity 
		real, intent(in)										:: dz,dx,dy !Spatial sampling 
		real, intent(inout)										:: oz,oy,ox
		integer, intent(out)									:: nz_out,nx_out,ny_out
		logical, intent(inout)									:: twoD
		!Internal variables
		integer													:: nz, nx, ny, iz, ix, iy
		real													:: oz_out, ox_out, oy_out

		!Reading the model dimensions
		nz = size(vel,1)
		nx = size(vel,2)
		ny = size(vel,3)
		

		!Checking the model dimensions and decide if 2D or 3D
		if ((nz .eq. 1) .or. (nx .eq. 1) .or. (ny .eq. 1)) then
			twoD = .true.

			!Allocating the wider velocity model (2D)
			allocate(vel_pad(nz+2*nv_pad,nx+2*nv_pad,ny))
			vel_pad = 0.

			!Putting the original velocity in the wider model
			!$OMP PARALLEL DO private(ix,iz)
		  	 do ix = 1,nx
		 	     do iz = 1,nz
				 vel_pad(iz+nv_pad,ix+nv_pad,:) = vel(iz,ix,:)
		 	     enddo
			enddo
			!$OMP END PARALLEL DO
	
			!Extending in the x
			!$OMP PARALLEL DO private(ix,iz)
			do ix = 1,nv_pad
			   do iz = 1,nz
				 vel_pad(iz+nv_pad,ix,:)=vel(iz,1,:)
				 vel_pad(iz+nv_pad,ix+nv_pad+nx,:)=vel(iz,nx,:)
			   enddo
			enddo
			!$OMP END PARALLEL DO

			!Extending in the z
			!$OMP PARALLEL DO private(ix,iz)
			do ix = 1,nx+2*nv_pad
			   do iz = 1,nv_pad
				 vel_pad(iz,ix,:)=vel_pad(nv_pad+1,ix,:)	
				 vel_pad(iz+nz+nv_pad,ix,:)=vel_pad(nz+nv_pad,ix,:)
			   enddo
			enddo
			!$OMP END PARALLEL DO

			!New dimensions
			nz_out = nz + 2*nv_pad
			nx_out = nx + 2*nv_pad
			ny_out = ny 
			oz_out = oz - dz*nv_pad; oz = oz_out
			ox_out = ox - dx*nv_pad; ox = ox_out	

		else
			twoD = .false.
			!Allocating the wider velocity model
			allocate(vel_pad(nz+2*nv_pad,nx+2*nv_pad,ny+2*nv_pad))
			vel_pad = 0.

			!Putting the original velocity in the wider model	
			!$OMP PARALLEL DO private(iy,ix,iz)
			do iy = 1,ny
			   do ix = 1,nx
			      do iz = 1,nz
				 vel_pad(iz+nv_pad,ix+nv_pad,iy+nv_pad) = vel(iz,ix,iy)
			      enddo
			   enddo	
			enddo
			!$OMP END PARALLEL DO

			!Extending in the x	
			!$OMP PARALLEL DO private(iy,ix,iz)
			do iy = 1,ny
			   do ix = 1,nv_pad
			      do iz = 1,nz
				 vel_pad(iz+nv_pad,ix,iy+nv_pad)=vel(iz,1,iy)
				 vel_pad(iz+nv_pad,ix+nv_pad+nx,iy+nv_pad)=vel(iz,nx,iy)
			      enddo
			   enddo
			enddo
			!$OMP END PARALLEL DO

			!Extending in the y
			!$OMP PARALLEL DO private(iy,ix,iz)
			do iy = 1,nv_pad
			   do ix = 1,nx+2*nv_pad
			      do iz = 1,nz
				 vel_pad(iz+nv_pad,ix,iy)=vel_pad(iz+nv_pad,ix,1+nv_pad)	
				 vel_pad(iz+nv_pad,ix,iy+nv_pad+ny)=vel_pad(iz+nv_pad,ix,ny+nv_pad)
			      enddo
			   enddo
			enddo
			!$OMP END PARALLEL DO

			!Extending in the z
			!$OMP PARALLEL DO private(iy,ix,iz)
			do iy = 1,ny+2*nv_pad
			   do ix = 1,nx+2*nv_pad
			      do iz = 1,nv_pad
				 vel_pad(iz,ix,iy)=vel_pad(1+nv_pad,ix,iy)
				 vel_pad(iz+nv_pad+nz,ix,iy)=vel_pad(nz+nv_pad,ix,iy)
			      enddo
			   enddo
			enddo	
			!$OMP END PARALLEL DO

			!New dimensions
			nz_out = nz + 2*nv_pad
			nx_out = nx + 2*nv_pad
			ny_out = ny + 2*nv_pad
			oz_out = oz - dz*nv_pad; oz = oz_out
			ox_out = ox - dx*nv_pad; ox = ox_out	
			oy_out = oy - dy*nv_pad; oy = oy_out
	
		endif
		
		deallocate(vel)
		return
	end subroutine Pad_velocity

end module Pad_vel
