!Module for solving one step of the acoustic wave equation
module Prop_step

  use weight
  use G
  use C5

  implicit none
  !Module variables
  integer, private								:: nz, nx, ny, nt
  real,allocatable,dimension(:,:,:),private		:: tmp1,tmp2

  contains


	!Initialization routine
	subroutine Prop_step_init(nz_in, nx_in, ny_in, nt_in, nv_pad_in, vel_in, dt_in, dz_in, dx_in, dy_in, twoD_in)
		
		!Input variables	
		integer, intent(in)						:: nz_in, nx_in, ny_in, nt_in, nv_pad_in
		real, dimension(:,:,:), pointer			:: vel_in
		real									:: dt_in, dz_in, dx_in, dy_in
		logical, intent(in)						:: twoD_in

		!Assigning the internal variables
		nz = nz_in
		nx = nx_in
		ny = ny_in
		nt = nt_in
		if(.not. allocated(tmp1)) allocate(tmp1(nz,nx,ny))
		if(.not. allocated(tmp2)) allocate(tmp2(nz,nx,ny))
 		
		!Initializing other modules
		call C5_init(nz, nx, ny, vel_in, dt_in, dz_in, dx_in, dy_in, twoD_in)	
		call weight_init(nz,nx,ny,dt_in,dz_in,dx_in,dy_in,nv_pad_in,vel_in,twoD_in)
  		call G_init(nz,nx,ny,dt_in,dz_in,dx_in,dy_in,nv_pad_in,twoD_in)
		

		return
	end subroutine Prop_step_init


	
	!Operator routine
	subroutine Prop_step_op(adj, pold, pcur, pnew)
	
		!Input-Output variables
		logical, intent(in)								:: adj
		real, dimension(nz,nx,ny), intent(inout)		:: pold, pcur, pnew
		!Internal variables

    	if(adj) then !Adjoint  
			call C5_op(.true.,.false.,tmp1,pcur)
			!Operator: (w-I)
			call weight_op(.true.,.false.,tmp2(:,:,:),pold)
			pnew(:,:,:) = tmp1(:,:,:) + tmp2(:,:,:) - pold(:,:,:) !Not Adding the source term
		else	     !Forward
			call C5_op(.false.,.false.,pcur,tmp1)
			!Operator: (w-I)
			call weight_op(.false.,.false.,pold,tmp2)
			pnew(:,:,:) = tmp1(:,:,:) + tmp2(:,:,:) - pold(:,:,:) !Not adding the source term
		endif

		return
	end subroutine Prop_step_op


	!Subroutine for deallocating variables
	subroutine Prop_step_close()
	  	call C5_close()
	  	if(allocated(tmp2)) deallocate(tmp2)
	  	if(allocated(tmp1)) deallocate(tmp1)
	  	call weight_close()
	end subroutine Prop_step_close


end module Prop_step

