!!$=head1 NAME
!!$
!!$Ricker - Program for computing a Ricker wavelet
!!$
!!$=head1 SYNOPSIS
!!$
!!$Ricker n1= o1=0.0 d1=0.004 freq= shift= >ricker.H
!!$
!!$=head1 DESCRIPTION
!!$
!!$Program that creates a Ricker wavelet for a certain central frequency and time shift
!!$
!!$
!!$=head1 PARAMETERS
!!$
!!$=over 3
!!$
!!$=item from history 
!!$
!!$
!!$=item from par
!!$
!!$B<integer> I<n1,o1,d1> - [~,0.0,0.004] Usual SEP axes parameters (samples,origin,sampling)
!!$
!!$B<real> I<freq> - Central frequency of the ricker
!!$
!!$B<real> I<shift> - Time shift of the wavelet
!!$
!!$
!!$=back
!!$
!!$=head1 USED MODULES 
!!$
!!$SEP, ricker_mod
!!$
!!$=head1 SEE ALSO
!!$
!!$Wavelet
!!$
!!$=cut
!%
program Ricker_main
	
	!Used modules 
  	use sep
  	use ricker_mod
  		
  	integer								:: nt
  	real								:: ot, dt, fd, td
  	real,dimension(:),allocatable       :: wavelet
	call sep_init (SOURCE)
	
  	call from_par("n1",nt)
  	call from_par("d1",dt)
  	call from_par("o1",ot,0.0)
  	call from_par("freq",fd)
  	call from_par("shift",td)
  	
  	
  	
  	allocate(wavelet(nt))
  	wavelet=0.
  	
  	
  	call ricker(nt,ot,dt,fd,td,wavelet)
  	call sep_write(wavelet)
  	call to_history("n1",nt)
  	call to_history("o1",ot)
  	call to_history("d1",dt)	
  	call to_history("label1","time [s]")
  	call to_history("label2","amplitude")
	
	
	
	
	call sep_close()
  	call exit (0)

end program Ricker_main