!Module for the scaling of the source (adjoint=forward)
!Model dimension (t)
!Data dimension (t)
!D(t) = Sou_scale(Model(t))
!Model(t) = Sou_scale*(D(t))

module Sou_scale


  implicit none
  !Module variables
  real, private										:: x, y, z, x0, y0, z0, dx, dy, dz, dt
  integer, private									:: nz, nx, ny
  logical, private									:: twoD
  real												:: dt2

  contains
	
	!Initialization routine
	subroutine Sou_scale_init(nz_in, nx_in, ny_in, z_in, x_in, y_in, z0_in, x0_in, y0_in, dz_in, dx_in, dy_in, dt_in,twoD_in)
		
		!Input variables	
		real, intent(in)				:: z_in, x_in, y_in, z0_in, x0_in, y0_in, dz_in, dx_in, dy_in, dt_in
		integer, intent(in)				:: nz_in, nx_in, ny_in
		logical, intent(in)				:: twoD_in

		!Assigning the internal variables
		x = x_in
		y = y_in
		z = z_in
		x0 = x0_in
		y0 = y0_in
		z0 = z0_in
		dx = dx_in
		dy = dy_in
		dz = dz_in
		nx = nx_in
		ny = ny_in
		nz = nz_in
		dt = dt_in
		twoD = twoD_in
		dt2= dt*dt


		return
	end subroutine Sou_scale_init

	!Operator routine
	subroutine Sou_scale_op(mm,vel)
	
		!Input-Output variables
		real, dimension(:), intent(inout)				:: mm !source (t)
		real, dimension(:,:,:), intent(in)				:: vel
		!Internal variables
		integer											:: zi, xi, yi
		real											:: zf, xf, yf
		real											:: wz, wx, wy




		
		if (twoD) then
			xf = 1. + (x-x0)/dx; xi = xf; wx = xf - xi
			zf = 1. + (z-z0)/dz; zi = zf; wz = zf - zi
			mm(:)       = mm(:)*dt2*(vel(zi,xi,1)*vel(zi,xi,1)*(1.-wz)*(1.-wx) &  
				      	+ vel(zi,xi+1,1)*vel(zi,xi+1,1)*(1.-wz)*(wx) &
						+ vel(zi+1,xi,1)*vel(zi+1,xi,1)*(wz)*(1.-wx)  &   
				  		+ vel(zi+1,xi+1,1)*vel(zi+1,xi+1,1)*(wz)*(wx) )  
		
		else
			xf = 1. + (x-x0)/dx; xi = xf; wx = xf - xi
			yf = 1. + (y-y0)/dy; yi = yf; wy = yf - yi
			zf = 1. + (z-z0)/dz; zi = zf; wz = zf - zi
			mm(:)       = mm(:)*dt2*(vel(zi,xi,yi)*vel(zi,xi,yi)*(1.-wz)*(1.-wx)*(1.-wy)  &       
				     	+ vel(zi,xi+1,yi)*vel(zi,xi+1,yi)*(1.-wz)*(wx)*(1.-wy)  &
				     	+ vel(zi,xi,yi+1)*vel(zi,xi,yi+1)*(1.-wz)*(1.-wx)*(wy)  &    
			 			+ vel(zi,xi+1,yi+1)*vel(zi,xi+1,yi+1)*(1.-wz)*(wx)*(wy) &   
			  		   	+ vel(zi+1,xi,yi)*vel(zi+1,xi,yi)*(wz)*(1.-wx)*(1.-wy)  &     
			   			+ vel(zi+1,xi+1,yi)*vel(zi+1,xi+1,yi)*(wz)*(wx)*(1.-wy) &    
			  	 		+ vel(zi+1,xi,yi+1)*vel(zi+1,xi,yi+1)*(wz)*(1.-wx)*(wy) &   
			 			+ vel(zi+1,xi+1,yi+1)*vel(zi+1,xi+1,yi+1)*(wz)*(wx)*(wy))	
		endif

		
		return
	end subroutine Sou_scale_op
	

	!Subroutine for deallocating variables
	subroutine Sou_scale_close()

	end subroutine Sou_scale_close

end module Sou_scale
