!Dot Product Test Module
module dot_mod

  implicit none

  contains

  subroutine dotp(operator, n_mm, n_dd)
    integer, intent(in)               		:: n_mm, n_dd
    real, dimension(2)		 			   	:: dot1, dot2

    interface
      subroutine operator(adj, add, mm, dd)
        logical,intent(in)  			:: adj,add
        real, dimension(:)  			:: mm, dd
      end subroutine operator
    end interface 

    !Declaring data and model dimensions
    real, dimension(n_mm)  				:: mm1,mm2
    real, dimension(n_dd)   			:: dd1, dd2

    write(0,*)"DOT PRODUCT TEST"
    mm1 = 0.; mm2 = 0.
    dd1 = 0.; dd2 = 0.
    write(0,*)"Generating random numbers"
    call random_number(mm1)
    call random_number(dd2)


    !dot product test with add = false
    write(0,*)"Computing forward add = false"
    call operator(.false.,.false.,mm1,dd1)
    dot1(1) = dot_product(dd1,dd2)
    write(0,*)"Computing adjoint add = false"
    call operator(.true.,.false.,mm2,dd2)
    dot1(2) = dot_product(mm1,mm2)

    !dot product test with add = true
    write(0,*)"Computing forward add = true"
    call operator(.false., .true., mm1, dd1)
    dot2(1) = dot_product(dd1, dd2)
    write(0,*)"Computing adjoint add = true"
    call operator(.true., .true., mm2, dd2)
    dot2(2) = dot_product(mm1, mm2)

    write(0,*) "Dot product with add = false: "
    write(0,*) "	Data space: ",dot1(1)
    write(0,*) "	Model space:",dot1(2)
    write(0,*) "	Error:      ",abs(dot1(1)-dot1(2))
    write(0,*) "	Relative:   ",abs(dot1(1)-dot1(2))/dot1(1)
    write(0,*) "Dot product with add = true: "
    write(0,*) "	Data space: ",dot2(1)
    write(0,*) "	Model space:",dot2(2)
    write(0,*) "	Error:      ",abs(dot2(1)-dot2(2))
    write(0,*) "	Relative:   ",abs(dot2(1)-dot2(2))/dot2(1)
    
    return
  end subroutine dotp
  
  subroutine dotp2(operator, n_mm, n_dd)
    integer, intent(in)               		:: n_mm, n_dd
    real, dimension(2)		 			   	:: dot1, dot2

    interface
      subroutine operator(adj, add, mm, dd)
        logical,intent(in)  			:: adj,add
        real, dimension(:)  			:: mm
        real							:: dd
      end subroutine operator
    end interface 

    !Declaring data and model dimensions
    real, dimension(n_mm)  				:: mm1,mm2
    real					  			:: dd1, dd2

    write(0,*)"DOT PRODUCT TEST"
    mm1 = 0.; mm2 = 0.
    dd1 = 0.; dd2 = 0.
    write(0,*)"Generating random numbers"
    call random_number(mm1)
    call random_number(dd2)


    !dot product test with add = false
    write(0,*)"Computing forward add = false"
    call operator(.false.,.false.,mm1,dd1)
    dot1(1) = dd1*dd2
    write(0,*)"Computing adjoint add = false"
    call operator(.true.,.false.,mm2,dd2)
    dot1(2) = dot_product(mm1,mm2)

    !dot product test with add = true
    write(0,*)"Computing forward add = true"
    call operator(.false., .true., mm1, dd1)
    dot2(1) = dd1*dd2
    write(0,*)"Computing adjoint add = true"
    call operator(.true., .true., mm2, dd2)
    dot2(2) = dot_product(mm1, mm2)

    write(0,*) "Dot product with add = false: "
    write(0,*) "	Data space: ",dot1(1)
    write(0,*) "	Model space:",dot1(2)
    write(0,*) "	Error:      ",abs(dot1(1)-dot1(2))
    write(0,*) "Dot product with add = true: "
    write(0,*) "	Data space: ",dot2(1)
    write(0,*) "	Model space:",dot2(2)
    write(0,*) "	Error:      ",abs(dot2(1)-dot2(2))
    
    return
  end subroutine dotp2


  subroutine dotp_cmplx(operator, n_mm, n_dd)
    integer, intent(in)               		:: n_mm, n_dd
    real, dimension(2)				    	:: dot1, dot2

    interface
      subroutine operator(adj, add, mm, dd)
        logical,intent(in)  			:: adj,add
        complex, dimension(:)  			:: mm, dd
      end subroutine operator
    end interface 

    !Declaring data and model dimensions
    complex, dimension(n_mm)  			:: mm1,mm2
    complex, dimension(n_dd)   			:: dd1, dd2
    real, dimension(n_mm)				:: r1_buff, r2_buff
    real, dimension(n_dd)				:: r3_buff, r4_buff

    write(0,*)"DOT PRODUCT TEST"
    mm1 = 0.; mm2 = 0.
    dd1 = 0.; dd2 = 0.
    r1_buff = 0.; r2_buff = 0.
    r3_buff = 0.; r4_buff = 0.
    write(0,*)"Generating random numbers"
    call random_number(r1_buff)
    call random_number(r2_buff)
    mm1 = dcmplx(r1_buff,r2_buff)
    call random_number(r3_buff)
    call random_number(r4_buff)
    dd2 = dcmplx(r3_buff,r4_buff)


    !dot product test with add = false
    write(0,*)"Computing forward add = false"
    call operator(.false.,.false.,mm1,dd1)
    dot1(1) = cabs(dot_product(dd1,dd2))
    write(0,*)"Computing adjoint add = false"
    call operator(.true.,.false.,mm2,dd2)
    dot1(2) = cabs(dot_product(mm1,mm2))


    !dot product test with add = true
    write(0,*)"Computing forward add = true"
    call operator(.false., .true., mm1, dd1)
    dot2(1) = cabs(dot_product(dd1, dd2))
    write(0,*)"Computing adjoint add = true"
    call operator(.true., .true., mm2, dd2)
    dot2(2) = cabs(dot_product(mm1, mm2))

    write(0,*) "Dot product with add = false: "
    write(0,*) "	Data space: ",dot1(1)
    write(0,*) "	Model space:",dot1(2)
    write(0,*) "	Error:      ",abs(dot1(1)-dot1(2))
    write(0,*) "Dot product with add = true: "
    write(0,*) "	Data space: ",dot2(1)
    write(0,*) "	Model space:",dot2(2)
    write(0,*) "	Error:      ",abs(dot2(1)-dot2(2))

    return
  end subroutine dotp_cmplx

end module dot_mod
