!Module that perfoms the 1D Fourier transformation
!Model dimension (t or x) (time or space)
!Data dimension (f)  (frequency (wavenumber))
!D(f) = fourier(Model(t or x))
!Model(t or x) = fourier*(D(f))
module fourier
	use sep

  implicit none
  include "fftw3.f"

  complex, dimension(:), allocatable, private  		:: in, out
  integer, private                       	      	:: n1
  integer, private                        	 	  	:: planf, planb
  real, private                      		 		:: sign1
  contains

  subroutine fourier_init(n1_in, sign_in)
    integer 	             	:: n1_in
    real		               	:: sign_in
    n1 = n1_in
    sign1 = sign_in  
    if(.not. allocated(in)) allocate(in(n1))
    if(.not. allocated(out)) allocate(out(n1))
    in = 0.
    out = 0.
    call sfftw_plan_dft_1d(planf, n1, in, out, FFTW_FORWARD, FFTW_ESTIMATE)
    call sfftw_plan_dft_1d(planb, n1, out, in, FFTW_BACKWARD, FFTW_ESTIMATE)
  end subroutine fourier_init
   
  subroutine fourier_call(adj,add,mm,dd)
    	logical,intent(in)   			:: adj,add
   		complex,dimension(:)  			:: mm,dd
    	call fourier_op(adj,add,mm,dd)
  end subroutine fourier_call

  subroutine fourier_op(adj, add, cmodel, cdata)
  ! Forward and Ajoint/Inverse Fourier Transform
    logical, intent(in)        		:: adj, add
    complex, dimension(:)    		:: cmodel, cdata
    if(adj) then !Adjoint
		if(.not. add) cmodel(:) = 0. !Zero-out the model
		out(1:(n1+1)/2) = cdata(n1-(n1+1)/2+1:n1)
        out((n1+1)/2+1:n1) = cdata(1:n1-(n1+1)/2)
        call sfftw_execute_dft(planb, out, in)
        cmodel(:) = cmodel(:) + sign1*in(:)/sqrt(float(n1))
    else !Forward
		if(.not. add) cdata(:) = 0. !Zero-out the data
        in = cmodel(:)
        call sfftw_execute_dft(planf, in, out)
        cdata(1:n1-(n1+1)/2) = cdata(1:n1-(n1+1)/2) + sign1*(out((n1+1)/2+1:n1))/sqrt(float(n1))
        cdata(n1-(n1+1)/2+1:n1) = cdata(n1-(n1+1)/2+1:n1) + sign1*(out(1:(n1+1)/2))/sqrt(float(n1))
    endif
  end subroutine fourier_op


  subroutine fourier_close()
    if(allocated(in))  deallocate(in)
    if(allocated(out)) deallocate(out)
    call sfftw_destroy_plan(planf)
    call sfftw_destroy_plan(planb)
  end subroutine fourier_close

end module
