!Module that resamples a signal from a coarse to a fine grid (linear)
!Model dimension (x1) (coarse grid)
!Data dimension (x2)  (fine grid)
!D(x2) = resample(Model(x1))
!Model(x2) = resample*(D(x1))
module resample

  implicit none
  !Module variables
  integer, private								:: nt, nw, desample
  real*8, dimension(:,:), allocatable, private	:: interp_filter

  contains
	
	!Initialization routine
	subroutine resample_init(nt_in, nw_in, desample_in)
		
		!Input variables	
		integer						:: nt_in, nw_in, desample_in,it2

		!Assigning the internal variables
		nt = nt_in
		nw = nw_in
		desample = desample_in
		!precompute the linear interpolation filter
  		allocate(interp_filter(desample+1,2))
  		do it2=1,desample+1
			interp_filter(it2,1)=1.-float((it2-1))/float(desample)
			interp_filter(it2,2)=1.-interp_filter(it2,1)
  		end do
  		interp_filter = interp_filter * sqrt(float(nw)/float(nt))

		return
	end subroutine resample_init

  	subroutine resample_call(adj,add,mm,dd)
    		logical,intent(in)   :: adj,add
   			real,dimension(:) 	 :: mm,dd
    		call resample_op(adj,add,mm,dd)
  	end subroutine resample_call

	
	!Operator routine
	subroutine resample_op(adj, add, mm, dd)
	
		!Input-Output variables
		logical, intent(in)				:: adj, add
		real, dimension(nw)				:: mm !coarse grid data (model)
		real, dimension(nt)				:: dd !fine grid data (data)
		!Internal variables
		integer							:: j, startind, endind

    	if(adj) then !Adjoint
        	if (.not. add) mm = 0. !Zero-out the model
			do j = 1, nw-1
				startind = desample*(j-1)+1
				endind = startind + desample -1
				mm(j) = mm(j) + sum(dd(startind:endind)*interp_filter(1:desample-1,1))
				mm(j+1) = mm(j+1) + sum(dd(startind:endind)*interp_filter(1:desample-1,2))
			enddo
			mm(nw) = mm(nw) + dd(nt) 
			

		else	     !Forward
			if(.not. add) dd = 0. !Zero-out the data
			do j = 1, nw-1
				startind = desample*(j-1)+1
				endind = startind + desample -1
				dd(startind:endind) = dd(startind:endind) + interp_filter(1:desample-1,1)*mm(j) + interp_filter(1:desample-1,2)*mm(j+1)
			enddo
			dd(nt) = dd(nt) + mm(nw)
		endif
		
		
		
		return
	end subroutine resample_op

  	subroutine resample_close()

  	end subroutine resample_close
	

end module resample
