!Module that resamples a signal from a coarse to a fine grid
!Model dimension (x1) (coarse grid)
!Data dimension (x2)  (fine grid)
!D(x2) = resample(Model(x1))
!Model(x2) = resample*(D(x1))
! Strange problem...routine not working if call before running dot product it does not pass the test...if it is called directly for dot product, it does pass the test
module resample_fourier

  use fourier

  implicit none
  !Module variables
  integer, private								:: nd, nm, nsample, start 	  !number of samples of the data and the model
  integer, private								:: nm2
  complex, dimension(:), allocatable, private	:: tmp, tmp1

  contains
	
	!Initialization routine
	subroutine resample_init(nm_in, nd_in)
		
		!Input variables	
		integer						:: nm_in, nd_in

		!Assigning the internal variables
		nm = nm_in
		nd = nd_in
		!Find next power of 2 of the model
		nm2 = 2.**(ceiling(log(real(nm))/log(2.)))
		if (nm2 .gt. nd) nm2  = nm !Use the original length if the next power of two is greater than nd
		if (nm .gt. nd) then 
			write(0,*)"Resample error: mismatching dimension (nm=",nm,"> nd=", nd,")"
			stop
		endif
		nsample = nd - nm2 ! number of samples to be padded
		if (mod(nm2,2).eq.0 .and. mod(nsample,2).ne.0) then
			start = nsample/2 + 1
		else
			start = (nsample+1)/2+1
		endif

		return
	end subroutine resample_init

  	subroutine resample_call(adj,add,mm,dd)
    		logical,intent(in)   :: adj,add
   			real,dimension(:) 	 :: mm,dd
    		call resample_op(adj,add,mm,dd)
  	end subroutine resample_call

	
	!Operator routine
	subroutine resample_op(adj, add, mm, dd)
	
		!Input-Output variables
		logical, intent(in)				:: adj, add
		real, dimension(nm)				:: mm !coarse grid data (model)
		real, dimension(nd)				:: dd !fine grid data (data)
		!Internal variables
		real							:: sign1
		parameter(sign1=1.)

    	if(adj) then !Adjoint
        	if (.not. add) mm = 0. !Zero-out the model
			allocate(tmp(nd),tmp1(nd))
			tmp = 0.; tmp1 = 0.
			tmp(:) = dd(:)
			!FFT
			call fourier_init(nd, sign1)
			call fourier_op(.false.,.false.,tmp,tmp1)
			call fourier_close()
			!Truncating to the desired length
			deallocate(tmp); allocate(tmp(nm2)); tmp(:) = 0.
			tmp(1:nm2) = tmp1(start:start+nm2)
			!tmp(:) = tmp1(1:nm2)!*(dsqrt(dble(nm2)/dble(nd)));!*(dsqrt(dble(nd)/dble(nm2))); !Normalization 
			deallocate(tmp1); allocate(tmp1(nm2)); tmp1(:) = 0.
			!IFFT
			call fourier_init(nm2, sign1)
			call fourier_op(.true.,.false.,tmp1,tmp)
			call fourier_close()
			mm(:) = mm(:) + real(tmp1(1:nm))
			deallocate(tmp1,tmp)
			

		else	     !Forward
			if(.not. add) dd = 0. !Zero-out the data
			allocate(tmp(nm2),tmp1(nm2))
			tmp = 0.; tmp1 = 0.
			tmp(1:nm) = mm(:)
			!FFT
			call fourier_init(nm2, sign1)
			call fourier_op(.false.,.false.,tmp,tmp1);
			call fourier_close()
			!Pad to the desired length
			deallocate(tmp); allocate(tmp(nd)); tmp(:) = 0.
			tmp(start:start+nm2) = tmp1(1:nm2)
			!tmp(1:nm2) = tmp1(:)!*(dsqrt(dble(nm2)/dble(nd))); !Normalization 
			deallocate(tmp1); allocate(tmp1(nd)); tmp1(:) = 0.
			!IFFT
			call fourier_init(nd, sign1)
			call fourier_op(.true.,.false.,tmp1,tmp)
			call fourier_close()
			dd(:) = dd(:) + real(tmp1(:))
			deallocate(tmp1,tmp)
		endif
		
		
		
		return
	end subroutine resample_op

  	subroutine resample_close()

  	end subroutine resample_close
	

end module resample
