!Module that computes the second order derivative of the input (Nm = Nd)
!Model dimension (t) 
!Data dimension (t)  
!D(t) = sec_deriv(Model(t))
!Model(t) = sec_deriv*(D(t))
module sec_deriv

  implicit none
  !Module variables
  integer, private				:: nd, nm 	  !number of samples of the data and the model
  real, private					:: dt
  real, dimension(2), private	:: coeff

  contains
	
	!Initialization routine
	subroutine sec_deriv_init(nm_in, dt_in)
		
		!Input variables	
		integer						:: nm_in
		real						:: dt_in

		!Assigning the internal variables
		nm = nm_in
		dt = dt_in
		nd = nm
		coeff(1) = 1.0/(dt*dt)
		coeff(2) = -2.0/(dt*dt)
		

		return
	end subroutine sec_deriv_init

  	subroutine sec_deriv_call(adj,add,mm,dd)
    	logical,intent(in)   	 :: adj,add
   		real,dimension(:) 		 :: mm,dd
    		call sec_deriv_op(adj,add,mm,dd)
  	end subroutine sec_deriv_call

	
	!Operator routine
	subroutine sec_deriv_op(adj, add, mm, dd)
	
		!Input-Output variables
		logical, intent(in)				:: adj, add
		real, dimension(nm)				:: mm !Regular grid data (model)
		real, dimension(nd)				:: dd !Irregular grid data (data)
		!Internal variables
		integer							:: id, im 

    		if(adj) then !Adjoint
        		if (.not. add) mm = 0. !Zero-out the model 
        		!DIR$ SIMD
        		do id = 2, nd-1
					mm(id-1) = mm(id-1) + coeff(1)*dd(id) 
					mm(id) 	 = mm(id)	+ coeff(2)*dd(id)
					mm(id+1) = mm(id+1) + coeff(1)*dd(id)
				enddo		
				
			else	     !Forward
				if(.not. add) dd = 0. !Zero-out the data
		        !DIR$ SIMD
				do id = 2, nd-1
					dd(id) = dd(id) + coeff(1)*mm(id-1) + coeff(2)*mm(id) + coeff(1)*mm(id+1)
				enddo				
			endif
		
		return
	end subroutine sec_deriv_op
	

end module sec_deriv




