!Module that computes the second order derivative of wavefields (Nm = Nd)
!Model dimension (z,x,y,t) 
!Data dimension (z,x,y,t) 
!D(z,x,y,t) = sec_deriv_field(Model(z,x,y,t))
!Model(z,x,y,t) = sec_deriv_field*(D(z,x,y,t))
module sec_deriv_field

  implicit none
  !Module variables
  integer, private				:: nt
  real*8, dimension(2),private	:: coeff

  contains
	
	!Initialization routine
	subroutine sec_deriv_field_init(nt_in, dt_in)
		
		!Input variables
		integer, intent(in)			:: nt_in
		real, intent(in)			:: dt_in
		real*8						:: dt

		!Assigning the internal variables
		nt = nt_in
		dt = dt_in
		coeff(1) = 1.0/(dt*dt)
		coeff(2) = -2.0/(dt*dt)

		return
	end subroutine sec_deriv_field_init

	
	!Operator routine
	subroutine sec_deriv_field_op(mm, dd, it)
		!Operator matrix = self-adjoint => mm = input always
		!|-2 1 0 0|
		!|1 -2 1 0|
		!|0 1 -2 1|
		!|0 0 1 -2|
	
		!Input-Output variables
		real, dimension(:,:,:,:), allocatable :: mm 
		real, dimension(:,:,:), allocatable	  :: dd 
		integer, intent(in)				  	  :: it
		!Internal variables
		
		if(it .eq. 1) then
			dd(:,:,:) = coeff(1)*mm(:,:,:,it+1) + coeff(2)*mm(:,:,:,it)
		elseif(it .eq. nt)then
			dd(:,:,:) = coeff(1)*mm(:,:,:,it-1) + coeff(2)*mm(:,:,:,it)
		else
			dd(:,:,:) = coeff(1)*(mm(:,:,:,it+1)+mm(:,:,:,it-1)) + coeff(2)*mm(:,:,:,it)
		endif
		
		return
	end subroutine sec_deriv_field_op
	

end module sec_deriv_field




